local _G = _G

-- addon name and namespace
local ADDON, NS = ...

local Addon = LibStub("AceAddon-3.0"):GetAddon(ADDON)

local AceEvent = LibStub:GetLibrary("AceEvent-3.0")

-- get translations
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale(ADDON)

-- the Outlets module
local Outlets = Addon:NewModule("Outlets", "AceEvent-3.0")

-- local functions
local tinsert           = table.insert
local pairs             = pairs

-- outlet
NS.OUTLET = "outlet"

local Outlet_Prototype = {}

-- outlet functions
function Outlet_Prototype:UpdateActive(event, name, active)
	if name ~= self:GetName() then
		return
	end

	if active then		
		self:RegisterMessage("MAGNETO_FORCE_CHANGED", "Update")
	else
		self:UnregisterMessage("MAGNETO_FORCE_CHANGED")
	end

	self:SetActive(active)
end

function Outlet_Prototype:Update()
	if not self:IsReady() then
		return
	end
	
	if self:IsActive() and
		-- Outlets:HasOutlet(self:GetName()) and
		-- Outlets:IsOutletActive(self:GetName()) and
		Addon:GetState() == Addon.TRACKING then
		self:Refresh()
	else
		self:Clear()
	end	
end

function Outlet_Prototype:SetActive(active)
	active = active and true or false
	
	if self._active ~= active then
		self._active = active
		
		self:HandleActiveChanged()
		
		self:Update()
	end
end

function Outlet_Prototype:IsActive()
	return self._active
end

function Outlet_Prototype:IsOperational()
	return self:IsReady() and self:IsActive()
end

function Outlet_Prototype:HandleActiveChanged()
	if not self:IsReady() then
		return
	end
	
	-- empty dummy, overwrite in specific implementation
end

function Outlet_Prototype:Refresh()
	if not self:IsReady() then
		return
	end
	
	-- empty dummy, overwrite in specific implementation
end

function Outlet_Prototype:Clear()
	if not self:IsReady() then
		return
	end
	
	-- empty dummy, overwrite in specific implementation
end

-- factory method for outlet creation
function NS:CreateOutlet(name, description)
	local module = NS:CreateModule(NS.OUTLET, name, description)

	module._active = false
	
	for k, v in pairs(Outlet_Prototype) do
		module[k] = v
	end
	
	AceEvent:Embed(module)

	module:RegisterMessage("MAGNETO_OUTLET_ACTIVE_CHANGED", "UpdateActive")
	
	return module
end

function NS:IsValidOutlet(module)
	if not NS:IsValidModule(module) or module:GetType() ~= NS.OUTLET then
		return false
	end

	for k, _ in pairs(Outlet_Prototype) do
		if not module[k] then
			return false
		end
	end
	
	return true
end

-- module data
local moduleData = {
	outlets     = {},
	activeNames = {},
}

-- module handling
function Outlets:OnInitialize()
	-- empty
end

function Outlets:OnEnable()
	self:RegisterOutlets()
end

function Outlets:OnDisable()
	self:UnregisterOutlets()
end

function Outlets:RegisterOutlets()
	for name, outlet in NS:IterateModulesByType(NS.OUTLET) do
		self:RegisterOutlet(outlet)
	end
end

function Outlets:RegisterOutlet(outlet)
	if not outlet then 
		self:Debug("RegisterOutlet: no outlet provided")
		return false
	end
	
	-- check interface
	if not NS:IsValidOutlet(outlet)	then
		self:Debug("RegisterOutlet: outlet doesn't implement interface")
		return false
	end
	
	local name = outlet:GetName()
	
	if not name or name == "" then
		self:Debug("RegisterOutlet: outlet doesn't provide name")
		return false
	end
	
	-- already registered
	if moduleData.outlets[name] then
		self:Debug("RegisterOutlet: outlet '" .. tostring(name) .. "' already registered")
		return false
	end
	
	-- register outlet
	moduleData.outlets[name] = outlet
	
	self:SendMessage("MAGNETO_OUTLET_REGISTERED_CHANGED", name, outlet, true)
	
	return true
end

function Outlets:UnregisterOutlets()
	for name, outlet in pairs(moduleData.outlets) do
		self:UnregisterOutlet(outlet)
	end
end

function Outlets:UnregisterOutlet(outlet)
	if not outlet or not outlet.GetName or not self:HasOutlet(outlet:GetName()) then 
		return
	end
	
	local name = outlet:GetName()
	
	-- unregister options
	self:UnregisterOutletOptions(outlet)
	
	-- deactivate
	self:SetOutletActive(name, false)
	
	-- unregister outlet
	moduleData.outlets[name] = nil
	
	self:SendMessage("MAGNETO_OUTLET_REGISTERED_CHANGED", name, outlet, false)
end

function Outlets:HasOutlet(name)
	return moduleData.outlets[name] and true or false
end

function Outlets:GetOutlet(name)
	return moduleData.outlets[name]
end

function Outlets:IsOutletOperational(name)
	return self:IsOutletReady(name) and self:IsOutletActive(name)
end

function Outlets:IsOutletReady(name)
	return moduleData.outlets[name] and moduleData.outlets[name]:IsReady() or false
end

function Outlets:IsOutletActive(name)
	return moduleData.activeNames[name] or false
end

function Outlets:SetOutletActive(name, active)
	if not self:HasOutlet(name) then
		return
	end
	
	active = active or false
	
	if self:IsOutletActive(name) == active then
		return
	end
	
	moduleData.activeNames[name] = active
	
	self:SendMessage("MAGNETO_OUTLET_ACTIVE_CHANGED", name, active)
end

function Outlets:ToggleOutletActive(name)
	self:SetOutletActive(name, not self:IsOutletActive(name) and true or false)
end

function Outlets:IterateOutlets()
	return pairs(moduleData.outlets)
end

-- test
function Outlets:Debug(msg)
	Addon:Debug("(Outlets) " .. msg)
end
